"""Miscellaneous general purpose functions."""

import numpy as np


def merge_df_into(df, newdf, key_cols):
    """
    Add newdf to df using key_cols and updating where necessary.

    New values always overwrite old values, even if NaN.
    """
    # create a copy of the original dataframe
    result = df.copy()
    # update existing rows and add new ones with combine_first
    # first, set the index to key_cols
    result.set_index(key_cols, inplace=True)
    newdf_indexed = newdf.set_index(key_cols)
    # update with combine_first (opposite direction from update())
    updated = newdf_indexed.combine_first(result)
    # add any missing columns from original df
    for col in result.columns:
        if col not in updated.columns:
            updated[col] = result[col]
    # reset index and return
    return updated.reset_index()


@property
def cols_to_underscore(self):
    """Replace spaces in column names of all dataframes with underscores."""
    for df in [self.r, self.s, self.sc]:
        df.columns = df.columns.str.replace(" ", "_")


def rm_under(str):
    """Replace underscore in a string with a space."""
    return str.replace("_", " ")


def mergedicts(original, update):
    """
    Merge two dicts into one.

    Parameters
    --
    original: first dict
    update: second dict
    """
    if update is not None:
        z = original | update
        return z
    else:
        return original


def findsmoothvariance(y, filtsig=0.1, nopts=False):
    """
    Estimate and then smooth the variance over replicates of data.

    Parameters
    --
    y: data - one column for each replicate
    filtsig: sets the size of the Gaussian filter used to smooth the variance
    nopts: if set, uses estimateerrorbar to estimate the variance
    """
    from scipy.ndimage import filters

    if y.ndim == 1:
        # one dimensional data
        v = estimateerrorbar(y, nopts) ** 2
    else:
        # multi-dimensional data
        v = np.var(y, 1)
    # apply Gaussian filter
    vs = filters.gaussian_filter1d(v, int(len(y) * filtsig))
    return vs


def estimateerrorbar(y, nopts=False):
    """
    Estimate measurement error for each data point.

    The errors found by calculating the standard deviation of the nopts
    data points closest to that data point.

    Parameters
    --
    y: data - one column for each replicate
    nopts: number of points used to estimate error bars
    """
    y = np.asarray(y)
    if y.ndim == 1:
        ebar = np.empty(len(y))
        if not nopts:
            nopts = np.round(0.1 * len(y))
        for i in range(len(y)):
            ebar[i] = np.std(np.sort(np.abs(y[i] - y))[:nopts])
        return ebar
    else:
        print("estimateerrorbar: works for 1-d arrays only.")


def natural_keys(text):
    """
    Find a set of keys.

    These keys can be used to sort a list of strings by the numeric
    values in the list entries.

    Eg sorted(list, key= natural_keys)

    Parameters
    ---------
    text: a string (but see above example)
    """
    import re

    def atof(text):
        try:
            retval = float(text)
        except ValueError:
            retval = text
        return retval

    return [
        atof(c)
        for c in re.split(r"[+-]?([0-9]+(?:[.][0-9]*)?|[.][0-9]+)", text)
    ]


def islistempty(ell):
    """
    Check if a list of lists is empty, e.g. [[]].

    Parameters
    --
    ell: list
    """
    if isinstance(ell, list):  # Is a list
        return all(map(islistempty, ell))
    return False  # Not a list


def makelist(c):
    """
    Ensure that a variable is a list.

    Parameters
    --
    c: variable to be made into a list
    """
    import numpy as np

    if isinstance(c, np.ndarray):
        return list(c)
    elif type(c) is not list:
        return [c]
    else:
        return c


def figs2pdf(savename):
    """Save all open figures to a pdf file."""
    import matplotlib.pyplot as plt
    from matplotlib.backends.backend_pdf import PdfPages

    if "." not in savename:
        savename += ".pdf"
    with PdfPages(savename) as pp:
        for i in plt.get_fignums():
            plt.figure(i)
            pp.savefig()



"""Function to parse data file generated by a Tecan fluorescence plate reader."""

import numpy as np


def parseTecan(dfd, rdict):
    """Parse a data frame from a Tecan plate reader into a dict."""
    # extract datatypes
    datatypes = (
        dfd[dfd.columns[0]]
        .iloc[
            np.nonzero(
                dfd[dfd.columns[0]]
                .str.startswith("Cycle Nr", na=False)
                .to_numpy()
            )[0]
            - 1
        ]
        .to_numpy()
    )
    # if only OD data measured
    if not isinstance(datatypes[0], str):
        datatypes = ["OD"]
    # add datatypes to rdict
    for datatype in datatypes:
        rdict[datatype] = []
    # extract times of measurements
    t = (
        dfd.loc[
            dfd[dfd.columns[0]].str.startswith("Time [s]", na=False),
            dfd.columns[1] :,
        ]
        .dropna(axis="columns")
        .astype("float")
        .mean()
        .to_numpy()
        / 3600.0
    )
    # deal with overflows
    df = dfd.replace(to_replace="OVER", value=np.nan)
    cols = df.columns
    # extract data and add to rdict
    df.index = df[cols[0]]
    for x in np.arange(1, 13):
        for y in "ABCDEFGH":
            well = y + str(x)
            if well in df.index:
                data = df.loc[well, cols[1] :].to_numpy(dtype="float")
                if data.ndim == 1:
                    data = data[None, :]
                for i, tv in enumerate(t):
                    rdict["time"].append(tv)
                    rdict["well"].append(well)
                    for k, datatype in enumerate(datatypes):
                        rdict[datatype].append(data[k, i])
    return rdict



"""Code for parsing data generated by a Tecan Sunrise platereader."""

import itertools

import numpy as np
import pandas as pd


def parseSunrise(dfd: pd.DataFrame, rdict: dict[str, list[float | str]]):
    """Parse a data frame from a Sunrise plate reader into a dict."""
    rdict["OD"] = []
    # extract times of measurements
    t = (
        np.array([float(str(ts).split("s")[0]) for ts in dfd.to_numpy()[0]])
        / 3600
    )
    t = t[~np.isnan(t)]
    # add well positions to dfd
    mask = pd.to_numeric(dfd[dfd.columns[-1]], errors="coerce").notna()
    dfd["well_positions"] = None
    value_iter = itertools.cycle(
        [
            f"{letter}{number}"
            for letter in "ABCDEFGH"
            for number in range(1, 13)
        ]
    )
    dfd.loc[mask, "well_positions"] = [
        next(value_iter) for _ in range(mask.sum())
    ]
    # extract data
    for x in np.arange(1, 13):
        for y in "ABCDEFGH":
            well = y + str(x)
            if well in dfd["well_positions"].values:
                data = dfd[dfd["well_positions"] == well].iloc[:, :-1].values
                data = data.reshape(data.size)
                for i, tv in enumerate(t):
                    rdict["time"].append(tv)
                    rdict["well"].append(well)
                    rdict["OD"].append(data[i])
    return rdict



"""Functions to import and export."""

from pathlib import Path

import numpy as np
import pandas as pd

import omniplate.admin as admin
import omniplate.clogger as clogger
import omniplate.omgenutils as gu
from omniplate.loaddata import sort_attributes


def savelog(self, fname=None):
    """
    Save log to file.

    Parameters
    --
    fname: string, optional
        The name of the file. If unspecified, the name of the experiment.

    Example
    -------
    >>> p.savelog()
    """
    # export log
    if fname:
        fnamepath = self.datadirpath / (fname + ".log")
    else:
        fnamepath = self.wdirpath / ("".join(self.allexperiments) + ".log")
    with fnamepath.open("w") as f:
        f.write(self.logstream.getvalue())
    print(f"Exported {fnamepath.name}.")


@clogger.log
def exportdf(self, fname=False, type="tsv", ldirec=False):
    """
    Export the dataframes.

    The exported data may either be tab-delimited or csv or json files.
    Dataframes for the (processed) raw data, for summary data, and for
    summary statistics and corrections, as well as a log file, are
    exported.

    Parameters
    ----------
    fname: string, optional
        The name used for the output files.
        If unspecified, the experiment or experiments is used.
    type: string
        The type of file for export, either 'json' or 'csv' or 'tsv'.
    ldirec: string, optional
        The directory to write. If False, the working directory is used.

    Examples
    --------
    >>> p.exportdf()
    >>> p.exportdf('processed', type= 'json')
    """
    if not fname:
        fname = "".join(self.allexperiments)
    if ldirec:
        ldirec = Path(ldirec)
        fullfname = str(ldirec / fname)
    else:
        fullfname = str(self.wdirpath / fname)
    # export data
    if type == "json":
        self.r.to_json(fullfname + "_r.json", orient="split")
        self.s.to_json(fullfname + "_s.json", orient="split")
        self.sc.to_json(fullfname + "_sc.json", orient="split")
    else:
        sep = "\t" if type == "tsv" else ","
        self.r.to_csv(fullfname + "_r." + type, sep=sep, index=False)
        self.s.to_csv(fullfname + "_s." + type, sep=sep, index=False)
        self.sc.to_csv(fullfname + "_sc." + type, sep=sep, index=False)
    print(f"Exported {Path(fullfname).stem}.")
    # export log to file
    self.savelog(fname)


def load_json_csv_tsv(rootname):
    """Load exported file and convert into dataframe."""
    experiment_name = Path(rootname).name
    try:
        # json files
        impdf = pd.read_json(f"{rootname}.json", orient="split")
        print(f"Imported {experiment_name}.json")
    except FileNotFoundError:
        try:
            # csv files
            impdf = pd.read_csv(f"{rootname}.csv", sep=",")
            print(f"Imported {experiment_name}.csv")
        except FileNotFoundError:
            try:
                # tsv files
                impdf = pd.read_csv(f"{rootname}.tsv", sep="\t")
                print(f"Imported {experiment_name}.tsv")
            except FileNotFoundError:
                print(
                    f"No file called {rootname}.json " "or .csv or .tsv found."
                )
                return
    # ensure all are imported as strings
    for var in ["experiment", "condition", "strain"]:
        impdf[var] = impdf[var].astype(str)
    return impdf


@clogger.log
def importdf(self, commonnames, info=True, sep="\t"):
    """
    Import dataframes saved as either json or csv or tsv files.

    Parameters
    ----------
    commonnames: list of strings
        A list of names for the files to be imported with one string for
        each experiment.

    Examples
    --------
    >>> p.importdf('Gal')
    >>> p.importdf(['Gal', 'Glu', 'Raf'])
    """
    commonnames = gu.makelist(commonnames)
    for commonname in commonnames:
        commonname = str(self.datadirpath / commonname)
        for df in ["r", "s", "sc"]:
            impdf = load_json_csv_tsv(f"{commonname}_{df}")
            # merge dataframes
            if hasattr(self, df):
                setattr(
                    self, df, pd.merge(getattr(self, df), impdf, how="outer")
                )
            else:
                setattr(self, df, impdf)
        print()
    # update attributes
    self.allexperiments = list(self.s.experiment.unique())
    self.allconditions.update(
        {
            e: list(self.s[self.s.experiment == e].condition.unique())
            for e in self.allexperiments
        }
    )
    self.allstrains.update(
        {
            e: list(self.s[self.s.experiment == e].strain.unique())
            for e in self.allexperiments
        }
    )
    for e in self.allexperiments:
        rdf = self.r[self.r.experiment == e]
        res = list((rdf.strain + " in " + rdf.condition).dropna().unique())
        res = [r for r in res if r != "nan in nan"]
        self.allstrainsconditions.update({e: res})
    # find datatypes with mean in self.s
    dtypdict = {}
    for e in self.allexperiments:
        # drop columns of NaNs - these are created by merge if a datatype
        # is in one experiment but not in another
        tdf = self.s[self.s.experiment == e].dropna(axis=1, how="all")
        dtypdict[e] = list(tdf.columns[tdf.columns.str.contains("mean")])
    self.datatypes.update(
        {e: [dt.split("mean_")[1] for dt in dtypdict[e]] for e in dtypdict}
    )
    # to reduce fragmentation
    self.r = self.r.copy()
    self.s = self.s.copy()
    self.sc = self.sc.copy()
    # initialise progress
    for e in self.allexperiments:
        admin.initialiseprogress(self, e)
    # display info on import
    if info:
        self.info
    # display warning if duplicates created
    if len(self.allexperiments) != np.unique(self.allexperiments).size:
        print(
            "\nLikely ERROR: data with the same experiment, condition, "
            "strain, and time now appears twice!!"
        )
    sort_attributes(self)


def save(self, name=None):
    """Remind user that save is undefined."""
    print("You probably mean exportdf.")



"""Functions to calculate summary statistics."""

import numpy as np
import pandas as pd
from scipy import integrate
from scipy.interpolate import interp1d
from scipy.signal import find_peaks
from scipy.stats import iqr


def statserr(d, **kwargs):
    """Use the half the interquartile range as an error."""
    return iqr(d) / 2


def find_smoothed_data_summary_stats(
    fitvar,
    derivname,
    statnames,
    nosamples,
    f,
    t,
    e,
    c,
    s,
    findareas,
    plotlocalmax,
    printpeakproperties,
    **kwargs,
):
    """
    Find smoothed data and summary statistics from fitderiv instance.

    Pass any kwargs to scipy's find_peaks via findlocalmaxderiv.

    Find summary statistics with bootstrapped errors including:
        - the maximal time derivative
        - the time at which the maximal time derivative occurs
        - the timescale found from inverting the maximal time derivative
        - the extreme values and range of the smoothed data
        - the lag time (the time when the tangent from the point with the
            maximal time derivative crosses a line parallel to the time-axis
            that passes through the first data point

    A summary statistic is given as the median of a distribution of the
    statistic calculated from time series sampled from the optimal Gaussian
    process.

    Its error is estimated as the interquartile range of this
    distribution.
    """
    # time-series for s dataframe
    df_for_s = pd.DataFrame(
        {
            "experiment": e,
            "condition": c,
            "strain": s,
            "time": t,
            f"smoothed_{fitvar}": f.f,
            f"smoothed_{fitvar}_err": np.sqrt(f.fvar),
            derivname: f.df,
            derivname + "_err": np.sqrt(f.dfvar),
            f"d/dt_{derivname}": f.ddf,
            f"d/dt_{derivname}_err": np.sqrt(f.ddfvar),
        }
    )
    # sample from GP
    fs, gs, hs = f.fitderivsample(nosamples)
    if "log_" in fitvar:
        orig_fitvar = "".join(fitvar.split("log_"))
        df_for_s[f"smoothed_{orig_fitvar}"] = np.exp(f.f)
        df_for_s[f"smoothed_{orig_fitvar}_err"] = np.sqrt(
            np.var(np.exp(fs), axis=1)
        )
    # stats for sc dataframe
    # min f
    min_f = fs[np.argmin(fs, 0), np.arange(nosamples)]
    # max f
    max_f = fs[np.argmax(fs, 0), np.arange(nosamples)]
    # range
    range_f = fs[np.argmax(fs, 0), :] - fs[np.argmin(fs, 0), :]
    # calculate df stats
    im = np.argmax(gs, 0)
    # max df
    max_df = gs[im, np.arange(nosamples)]
    # time of max df
    t_max_df = np.array([t[i] for i in im])
    # inverse max df
    dt = np.log(2) / max_df
    # lag time
    lagtime = (
        t_max_df
        + (fs[0, np.arange(nosamples)] - fs[im, np.arange(nosamples)]) / max_df
    )
    # find local maximal derivative
    peak_coords, da, d = findlocalmaxderiv(
        f, gs, printpeakproperties, **kwargs
    )
    # find area under df/dt vs f and area under f vs t
    if findareas:
        adff, andff, aft, anft = findareasunder(t, fs, gs)
    else:
        adff, andff, aft, anft = np.nan, np.nan, np.nan, np.nan
    # store results
    dict_for_sc = {
        "experiment": e,
        "condition": c,
        "strain": s,
        f"local_max_{derivname}": peak_coords[1],
        f"local_max_{derivname}_err": statserr(da),
        f"time_of_local_max_{derivname}": peak_coords[0],
        f"time_of_local_max_{derivname}_err": statserr(dt),
        # area under df/dt vs f
        f"area_under_{derivname}_vs_{fitvar}": np.median(adff),
        f"area_under_{derivname}_vs_{fitvar}_err": statserr(adff),
        f"normalised_area_under_{derivname}_vs_{fitvar}": np.median(andff),
        f"normalised_area_under_{derivname}_vs_{fitvar}_err": statserr(andff),
        # area under f vs t
        f"area_under_{fitvar}": np.median(aft),
        f"area_under_{fitvar}_err": statserr(aft),
        f"normalised_area_under_{fitvar}": np.median(anft),
        f"normalised_area_under_{fitvar}_err": statserr(anft),
    }
    # add statnames stats
    for stname, st in zip(
        statnames, [min_f, max_f, range_f, max_df, t_max_df, dt, lagtime]
    ):
        dict_for_sc[stname] = np.median(st)
        dict_for_sc[stname + "_err"] = statserr(st)
    if not plotlocalmax:
        peak_coords = None
    return df_for_s, dict_for_sc, peak_coords


def findlocalmaxderiv(f, gs, printpeakproperties, **kwargs):
    """
    Find the greatest local maxima in the derivative.

    Check the derivative for local maxima and so find the local maximum
    with the highest derivative using samples, gs, of df.

    The keyword variables kwargs are passed to scipy's find_peaks.
    """
    # find peaks in mean df
    lpksmn, lpksmndict = find_peaks(f.df, **kwargs)
    if np.any(lpksmn):
        peak_coords = (
            f.t[lpksmn[np.argmax(f.df[lpksmn])]],
            f.df[lpksmn[np.argmax(f.df[lpksmn])]],
        )
        if printpeakproperties:
            # display properties of peaks
            print("Peak properties\n---")
            for prop in lpksmndict:
                print(f"{prop:15s}", lpksmndict[prop])
        # da: samples of local max df
        # dt: samples of time of local max df
        da, dt = [], []
        # find peaks of sampled df
        for gsample in np.transpose(gs):
            tpks = find_peaks(gsample, **kwargs)[0]
            if np.any(tpks):
                da.append(np.max(gsample[tpks]))
                dt.append(f.t[tpks[np.argmax(gsample[tpks])]])
        return peak_coords, da, dt
    else:
        # mean df has no peaks
        return (np.nan, np.nan), np.nan, np.nan


def findareasunder(t, fs, gs):
    """
    Find areas.

    Given samples of f, as arguments fs, and of df, as arguments gs,
    find the area under df/dt vs f and the area under f vs t.
    """
    adff, andff, aft, anft = [], [], [], []
    for fsample, gsample in zip(np.transpose(fs), np.transpose(gs)):
        # area under df vs f: integrand has f as x and df as y
        def integrand(x):
            return interp1d(fsample, gsample)(x)

        iresult = integrate.quad(
            integrand,
            np.min(fsample),
            np.max(fsample),
            limit=100,
            full_output=1,
        )[0]
        adff.append(iresult)
        andff.append(iresult / (np.max(fsample) - np.min(fsample)))

        # area under f vs t: integrand has t as x and f as y
        def integrand(x):
            return interp1d(t, fsample)(x)

        iresult = integrate.quad(
            integrand, np.min(t), np.max(t), limit=100, full_output=1
        )[0]
        aft.append(iresult)
        anft.append(iresult / (np.max(t) - np.min(t)))
    return adff, andff, aft, anft



"""
Functions for performing corrections.

For non-linearity in the OD, for the fluorescence of the media,
and for autofluorescence.
"""

import importlib.resources as import_files
import re

import gaussianprocessderivatives as gp
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from nunchaku import Nunchaku
from scipy.interpolate import interp1d
from scipy.optimize import minimize_scalar
from statsmodels.nonparametric.smoothers_lowess import lowess

import omniplate.admin as admin
import omniplate.clogger as clogger
import omniplate.omerrors as errors
import omniplate.omgenutils as gu
import omniplate.sunder as sunder
from omniplate.correctauto_bayesian import correctauto_bayesian
from omniplate.omfitderiv import preprocess_data, runfitderiv


@clogger.log
def correctOD(
    self,
    figs=True,
    bd=None,
    gp_results=False,
    ODfname=None,
    odmatch_min=0.1,
    max_nunchaku_segments=4,
    correctformedia=True,
    frac=0.33,
    null_dict=None,
    experiments="all",
    experimentincludes=False,
    experimentexcludes=False,
    conditions="all",
    conditionincludes=False,
    conditionexcludes=False,
):
    """
    Correct for the non-linear relationship between OD and cell number.

    Requires a set of dilution data set, with the default being haploid
    yeast growing in glucose.

    An alternative can be loaded from a file - a txt file of two columns
    with OD specified in the first column and the dilution factor specified
    in the second.

    Parameters
    ---------
    figs: boolean, optional
        If True, a plot of the fit to the dilution data is produced.
    bd: dictionary, optional
        Specifies the limits on the hyperparameters for the Gaussian
        process.
        For example, bd= {0: [-1, 4], 2: [2, 6]})
        sets confines the first hyperparameter to be between 1e-1 and 1e^4
        and confines the third hyperparameter between 1e2 and 1e6.
    gp_results: boolean, optional
        If True, show the results of fitting the Gaussian process
    ODfname: string, optional
        The name of the file with the dilution data used to correct OD for
        its non-linear dependence on numbers of cells. If unspecified, data
        for haploid budding yeast growing in glucose is used.
    odmatch_min: float, optional
        An expected minimal value of the OD up to which there is a linear
        scaling of the OD with cell numbers.
    max_nunchaku_segments: int, optional
        The number of segments used by nunchaku when finding the inital
        region where the relationship between the dilution factor and OD
        is linear.
    correctformedia: boolean, optional
        If True (default), correct OD for background levels from media.
    frac: float
        The fraction of the data used for smoothing the media OD via lowess.
        Used to correct OD for the background OD of the media.
    null_dict: dict[str, list[str]], optional
        A dictionary specifying which Null conditions should be used to correct
        other conditions.
        For example:
            null_dict= {"2% glu": ["all"]}
        means that all conditions should be corrected using "Null in 2% glu";
            null_dict= {"2% glu": ["2% glu 0.1 mg/ml", "2% glu 0.2 mg/ml"],
                        "2% gal": ["2% gal 0.1 mg/ml"]}
        means that "2% glu 0.1 mg/ml" and "2% glu 0.2 mg/ml" should be
        corrected with "Null in 2% glu" and that "2% gal 0.1 mg/ml" should
        be corrected with "Null in 2% gal".
    experiments: string or list of strings
        The experiments to include.
    conditions: string or list of strings
        The conditions to include.
    experimentincludes: string, optional
        Selects only experiments that include the specified string in their
        name.
    experimentexcludes: string, optional
        Ignores experiments that include the specified string in their
        name.
    conditionincludes: string, optional
        Selects only conditions that include the specified string in their
        name.
    conditionexcludes: string, optional
        Ignores conditions that include the specified string in their name.

    Examples
    -------
    >>> p.correctOD()
    >>> p.correctOD(figs= False)
    """
    exps = sunder.getset(
        self,
        experiments,
        experimentincludes,
        experimentexcludes,
        "experiment",
        nonull=True,
    )
    cons = sunder.getset(
        self,
        conditions,
        conditionincludes,
        conditionexcludes,
        "condition",
        nonull=True,
        nomedia=False,
    )
    # fit dilution data
    gc, odmatch = findODcorrection(
        datadirpath=self.datadirpath,
        ODfname=ODfname,
        figs=figs,
        bd=bd,
        gp_results=gp_results,
        odmatch_min=odmatch_min,
        max_nunchaku_segments=max_nunchaku_segments,
    )
    print()
    # correct ODs
    for e in exps:
        for c in cons:
            if correctformedia:
                correctODformedia(
                    self,
                    figs=figs,
                    frac=frac,
                    experiments=e,
                    conditions=c,
                    null_dict=null_dict,
                )
            # correct all wells
            r_data = self.r.query(
                "experiment == @e and condition == @c"
            ).OD.to_numpy()
            gc.batchpredict(r_data)
            # leave small ODs unchanged
            new_r = gc.f
            new_r[r_data < odmatch] = r_data[r_data < odmatch]
            # update r data frame
            self.r.loc[
                (self.r.experiment == e) & (self.r.condition == c),
                "OD",
            ] = new_r
    if self.progress["negativevalues"][e]:
        print(
            "Warning: correcting OD for media has created "
            f"negative values in {e} for"
        )
        # ignore final newline
        print(self.progress["negativevalues"][e][:-1])
    # update s dataframe
    admin.update_s(self)


def findODcorrection(
    datadirpath,
    ODfname,
    figs,
    bd,
    gp_results,
    odmatch_min,
    max_nunchaku_segments,
):
    """
    Determine a function to correct OD.

    Use a Gaussian process to fit serial dilution data to correct for
    non-linearities in the relationship between OD and cell density.

    The data are either loaded from file ODfname or the default
    data for haploid yeast growing in glucose are used.
    """
    print("Fitting dilution data for OD correction for non-linearities.")
    if ODfname is not None:
        try:
            od_df = pd.read_csv(
                str(datadirpath / ODfname),
                sep=None,
                engine="python",
                header=None,
            )
            print(f"Using {ODfname}")
            od_data = od_df.to_numpy()
            od, dilfac = od_data[:, 0], od_data[:, 1]
        except (FileNotFoundError, OSError):
            raise errors.FileNotFound(str(datadirpath / ODfname))
    else:
        print("Using default data.")
        fname = "dilution_data_xiao.tsv"
        od, dilfac = read_default_dilution_data(fname)
    od, dilfac = arrange_into_replicates(od, dilfac)
    # run nunchaku
    X = np.mean(dilfac, 1)
    nc = Nunchaku(X, od.T, estimate_err=True, prior=[-5, 5])
    num_regions, _ = nc.get_number(max_nunchaku_segments)
    bds, _ = nc.get_iboundaries(num_regions)
    # find linear region, which starts from origin
    odmatch_pts = np.mean(od, 1)[bds]
    # pick end point with OD at least equal to odmatch_min
    ipick = np.where(odmatch_pts > odmatch_min)[0][0]
    odmatch = odmatch_pts[ipick]
    dilfacmatch = X[bds[ipick]]
    # process data
    dilfac = dilfac.flatten()[np.argsort(od.flatten())]
    od = np.sort(od.flatten())
    # rescale so that OD and dilfac match
    y = dilfac / dilfacmatch * odmatch
    # set up Gaussian process
    bds = {0: (-4, 4), 1: (-4, 4), 2: (-3, -1)}
    # find bounds
    if bd is not None:
        bds = gu.mergedicts(original=bds, update=bd)
    gc = gp.maternGP(bds, od, y)
    # run Gaussian process
    gc.findhyperparameters(noruns=5, exitearly=True, quiet=True)
    if gp_results:
        gc.results()
    gc.predict(od)
    if figs:
        plt.figure()
        gc.sketch(".")
        plt.plot(odmatch, odmatch, "bs")
        plt.grid(True)
        plt.xlabel("OD")
        plt.ylabel("corrected OD (relative cell numbers)")
        if ODfname:
            plt.title("Fitting " + ODfname)
        else:
            plt.title("for haploid budding yeast in glucose")
        plt.show(block=False)
    return gc, odmatch


def read_default_dilution_data(fname):
    """Import default dilution data."""
    d = import_files.read_text("omniplate", fname)
    res = np.array(re.split(r"\n|\t", d)[:-1]).astype(float)
    od, dilfac = res[::2], res[1::2]
    if fname == "dilution_data_xiao.tsv":
        # missing replicate - use mean of existing ones
        dilfac = np.insert(dilfac, 0, dilfac[0])
        od = np.insert(od, 0, np.mean(od[:2]))
    else:
        raise ValueError("Dilution data unrecognised.")
    return od, dilfac


def arrange_into_replicates(od, dilfac):
    """Rearrange so that data from each replicate is in a column."""
    udilfac, indices, counts = np.unique(
        dilfac, return_inverse=True, return_counts=True
    )
    ucounts = np.unique(counts)
    if len(ucounts) == 1:
        noreps = np.unique(counts)[0]
        dilfac_reps = np.tile(np.atleast_2d(udilfac).T, noreps)
        od_reps = np.array(
            [od[indices == i] for i in range(udilfac.size)]
        ).reshape((udilfac.size, noreps))
        return od_reps, dilfac_reps
    else:
        raise ValueError(
            "There are inconsistent numbers of replicates"
            " in the OD correction data."
        )


@clogger.log
def correctODformedia(
    self,
    figs=False,
    frac=0.33,
    null_dict=None,
    experiments="all",
    experimentincludes=False,
    experimentexcludes=False,
    conditions="all",
    conditionincludes=False,
    conditionexcludes=False,
):
    """
    Correct OD or fluorescence for that of the media.

    Use data from wells marked Null.

    Use lowess to smooth measurements from all Null wells and subtract
    this smoothed time series from the raw data.
    """
    exps = sunder.getset(
        self,
        experiments,
        experimentincludes,
        experimentexcludes,
        "experiment",
        nonull=True,
    )
    cons = sunder.getset(
        self,
        conditions,
        conditionincludes,
        conditionexcludes,
        "condition",
        nonull=True,
        nomedia=False,
    )
    for e in exps:
        for c in cons:
            if c in self.allconditions[e]:
                print(f"{e} - {c}: Correcting OD for the OD of the medium.")
                negvalues = find_Null_and_correct(
                    self,
                    df=self.r,
                    dtype="OD",
                    e=e,
                    c=c,
                    figs=figs,
                    frac=frac,
                    null_dict=null_dict,
                )
                if negvalues is not None:
                    if not self.progress["negativevalues"][e]:
                        self.progress["negativevalues"][e] = negvalues
                    else:
                        self.progress["negativevalues"][e] += negvalues
    # update s dataframe
    admin.update_s(self)


def find_Null_and_correct(self, df, dtype, e, c, figs, frac, null_dict):
    """Find data for Null strain and pass df to perform_media_correction."""
    null_e, null_c = e, c
    if null_dict:
        for medium, conditions in null_dict.items():
            if conditions == ["all"] or null_c in conditions:
                null_c = medium
                break
    null_df = self.r[
        (self.r.experiment == null_e)
        & (self.r.condition == null_c)
        & (self.r.strain == "Null")
    ]
    if null_df.empty:
        print(f"{e}: No Null strain found for {c}.")
    else:
        negvalues = perform_media_correction(
            null_df=null_df,
            df=df,
            dtype=dtype,
            experiment=e,
            condition=c,
            figs=figs,
            frac=frac,
        )
        return negvalues


def perform_media_correction(
    null_df, df, dtype, experiment, condition, figs, frac
):
    """
    Correct data of type dtype for any signal from the media.

    Data for the Null strain is in null_df; data to be ovewritten
    is in df.

    Use lowess to smooth over time the media data from the Null
    wells and subtract the smoothed values from the data.
    """
    t, null_data = null_df["time"].to_numpy(), null_df[dtype].to_numpy()
    if ~np.any(null_data[~np.isnan(null_data)]):
        # all data is NaN
        return
    # find correction
    res = lowess(null_data, t, frac=frac)
    correctionfn = interp1d(
        res[:, 0],
        res[:, 1],
        fill_value=(res[0, 1], res[-1, 1]),
        bounds_error=False,
    )
    if figs:
        plt.figure()
        plt.plot(t, null_data, "ro", res[:, 0], res[:, 1], "b-")
        plt.xlabel("time (hours)")
        plt.title(
            f"{experiment}: media correction for {dtype} in {condition}."
        )
        plt.show(block=False)
    # perform correction to data in df
    choose = (df.experiment == experiment) & (df.condition == condition)
    df.loc[choose, dtype] = df[choose][dtype] - correctionfn(
        df[choose]["time"]
    )
    # check for any negative values
    negvalues = ""
    for s in np.unique(df[choose]["strain"][df[choose][dtype] < 0]):
        if s != "Null":
            wstr = f"\t{dtype}: {s} in {condition} for wells "
            for well in np.unique(
                df[choose][df[choose].strain == s]["well"][
                    df[choose][dtype] < 0
                ]
            ):
                wstr += f"{well}, "
            wstr = wstr[:-2] + "\n"
            negvalues += wstr
    return negvalues


@clogger.log
def correctauto(
    self,
    f="GFP",
    refstrain="WT",
    noprocessors=1,
    options=None,
    experiments="all",
    experimentincludes=False,
    experimentexcludes=False,
    conditions="all",
    conditionincludes=False,
    conditionexcludes=False,
    strains="all",
    strainincludes=False,
    strainexcludes=False,
):
    """
    Correct fluorescence for autofluorescence.

    The correction is made using the fluorescence of an untagged
    reference strain.

    Arguments
    --
    f: string
        The fluorescence measurements, typically either 'mCherry' or
        'GFP'.
    refstrain: string
        The reference strain used to estimate autofluorescence.
    noprocessors: int (default: 1)
        The number of processors to use. Numpy has built-in parallelisation
        and using one processor may be fastest.
    options: None or Dict
       The keys of options are:
            bd: dict (default: None)
                Specifies the bounds on the hyperparameters for the
                Gaussian process applied to the logarithm of the inferred
                fluorescence.
                e.g. {2: (-2, 0)}.
            figs: boolean (default: True)
                If True, use omplot.inspect to show raw data and inferred
                fluorescence per cell.
            noboots: int (default: 10)
                The number of bootstrapped data sets to fit. Larger numbers
                give better estimates of errors.
            flcvfn: str (default: "matern")
                The covariance function to use for the Gaussian process
                that smooths the inferred fluorescence.
            nosamples: int (default: 1000)
                The number of samples taken to estimate errors from the
                Gaussian process used to smooth the inferred fluorescence.
            maxdatapts: int or None (default: None)
                The maximum number of data points to allow in the Gaussian
                process smoothing before subsampling.
            fitderiv_figs: boolean (default: False)
                If True, show the results of the Gaussian process smoothing.
    experiments: string or list of strings
        The experiments to include.
    conditions: string or list of strings
        The conditions to include.
    strains: string or list of strings
        The strains to include.
    experimentincludes: string, optional
        Selects only experiments that include the specified string in
        their name.
    experimentexcludes: string, optional
        Ignores experiments that include the specified string in their
        name.
    conditionincludes: string, optional
        Selects only conditions that include the specified string in their
        name.
    conditionexcludes: string, optional
        Ignores conditions that include the specified string in their name.
    strainincludes: string, optional
        Selects only strains that include the specified string in their
        name.
    strainexcludes: string, optional
        Ignores strains that include the specified string in their name.

    Examples
    --------
    >>> p.correctauto('GFP', options = {"figs": False, "bd": {2: (-1,3)}})
    >>> p.correctauto('mCherry', refstrain= 'BY4741')
    """
    f = gu.makelist(f)
    if len(f) != 1:
        print("Error: correctauto uses only one fluorescence measurement.")
        return
    # correct for autofluorescence
    default_options = {
        "bd": None,
        "figs": True,
        "noboots": 10,
        "flcvfn": "matern",
        "nosamples": 1000,
        "maxdatapts": None,
        "fitderiv_figs": False,
    }
    if options is None:
        options = default_options
    else:
        options = default_options | options
    print(f"\nCorrecting autofluorescence using {refstrain} as the reference.")
    correctauto_bayesian(
        self,
        f,
        refstrain,
        noprocessors,
        experiments,
        experimentincludes,
        experimentexcludes,
        conditions,
        conditionincludes,
        conditionexcludes,
        strains,
        strainincludes,
        strainexcludes,
        options,
    )


@clogger.log
def correctauto_l(
    self,
    f=["GFP", "AutoFL"],
    refstrain="WT",
    method="default",
    options=None,
    experiments="all",
    experimentincludes=False,
    experimentexcludes=False,
    conditions="all",
    conditionincludes=False,
    conditionexcludes=False,
    strains="all",
    strainincludes=False,
    strainexcludes=False,
):
    """
    Correct fluorescence for autofluorescence.

    The correction is made using the fluorescence of an untagged
    reference strain.

    The reference strain is used to estimate the autofluorescence via
    either the method of Lichten et al., 2014, where measurements of
    fluorescence at two wavelengths is required, or by using the
    fluorescence of the reference strain interpolated to the OD of the
    strain of interest (Berthoumieux et al., 2013).

    Using two measurements of fluorescence is thought to be more accurate,
    particularly for low fluorescence measurements (Mihalcescu et al.,
    2015).

    Arguments
    --
    f: string or list of strings
        The fluorescence measurements, typically either ['mCherry'] or
        ['GFP', 'AutoFL'].
    refstrain: string
        The reference strain used to estimate autofluorescence.
    method: string
        Either "default" or "bayesian".
    options: None or Dict
        If method = "default", the keys of options are:
            figs: boolean
                If True, display plots showing the fits to the reference
                strain's fluorescence.
            useGPs: boolean
                If True, use Gaussian processes to generate extra samples
                from the replicates. Recommended, particularly if there
                are only a few replicates, but slower.
            flcvfn: str, optional
                The covariance function to use for the Gaussian process
                applied to the logarithm of the fluorescence if useGPs=True.
            bd: dict, optional
                Specifies the bounds on the hyperparameters for the
                Gaussian process applied to the logarithm of the
                fluorescence,
                e.g. {2: (-2, 0)}.
            nosamples: int, optional
                The number of samples to take when using Gaussian processes.
            maxdatapts: int, optional
                The maximum number of data points to use for the Gaussian
                process. Too many data points, over 1500, can be slow.
        with two fluorescence measurements, you can also specify
            frac: float, optional
                The fraction of the data used for smoothing via statmodels'
                lowess.
            null_dict: dict[str, list[str]], optional
                A dictionary specifying which Null conditions should be used to correct
                other conditions.
                For example:
                    null_dict= {"2% glu": ["all"]}
                means that all conditions should be corrected using "Null in 2% glu";
                    null_dict= {"2% glu": ["2% glu 0.1 mg/ml", "2% glu 0.2 mg/ml"],
                                "2% gal": ["2% gal 0.1 mg/ml"]}
                means that "2% glu 0.1 mg/ml" and "2% glu 0.2 mg/ml" should be
                corrected with "Null in 2% glu" and that "2% gal 0.1 mg/ml" should
                be corrected with "Null in 2% gal".
    experiments: string or list of strings
        The experiments to include.
    conditions: string or list of strings
        The conditions to include.
    strains: string or list of strings
        The strains to include.
    experimentincludes: string, optional
        Selects only experiments that include the specified string in
        their name.
    experimentexcludes: string, optional
        Ignores experiments that include the specified string in their
        name.
    conditionincludes: string, optional
        Selects only conditions that include the specified string in their
        name.
    conditionexcludes: string, optional
        Ignores conditions that include the specified string in their name.
    strainincludes: string, optional
        Selects only strains that include the specified string in their
        name.
    strainexcludes: string, optional
        Ignores strains that include the specified string in their name.

    Examples
    --------
    To correct data with one type of fluorescence measurement, use:

    >>> p.correctauto('GFP', options = {"figs" : False})
    >>> p.correctauto('mCherry', refstrain= 'BY4741')

    To correct data with two types of fluorescence measurement, use:

    >>> p.correctauto(['GFP', 'AutoFL'], options= {"useGPs" : True})
    >>> p.correctauto(['GFP', 'AutoFL'], refstrain= 'WT')
    >>> p.correctauto(['GFP', 'AutoFL'], refstrain= 'WT', method = "bayesian")

    References
    ----------
    S Berthoumieux, H De Jong, G Baptist, C Pinel, C Ranquet, D Ropers,
    J Geiselmann (2013).
    Shared control of gene expression in bacteria by transcription factors
    and global physiology of the cell.
    Mol Syst Biol, 9, 634.

    CA Lichten, R White, IB Clark, PS Swain (2014).
    Unmixing of fluorescence spectra to resolve quantitative time-series
    measurements of gene expression in plate readers.
    BMC Biotech, 14, 1-11.

    I Mihalcescu, MVM Gateau, B Chelli, C Pinel, JL Ravanat (2015).
    Green autofluorescence, a double edged monitoring tool for bacterial
    growth and activity in micro-plates.
    Phys Biol, 12, 066016.
    """
    f = gu.makelist(f)
    exps, cons, strains = sunder.getall(
        self,
        experiments,
        experimentincludes,
        experimentexcludes,
        conditions,
        conditionincludes,
        conditionexcludes,
        strains,
        strainincludes,
        strainexcludes,
        nonull=True,
    )
    # correct for autofluorescence
    if method == "default" and len(f) == 2:
        default_options = {
            "figs": True,
            "useGPs": True,
            "flcvfn": "matern",
            "bd": None,
            "nosamples": 1000,
            "maxdatapts": None,
            "frac": 0.33,
            "null_dict": None,
        }
    elif method == "default" and len(f) == 1:
        default_options = {
            "figs": True,
            "useGPs": True,
            "flcvfn": "matern",
            "bd": None,
            "nosamples": 1000,
            "maxdatapts": None,
        }
    if options is None:
        options = default_options
    else:
        options = default_options | options
    checking_error = correctauto_checks_l(
        self, f, method, options, exps, cons, strains
    )
    if checking_error:
        print("correctauto: failed checks.")
        return
    else:
        print(
            f"\nCorrecting autofluorescence using {refstrain} as the"
            " reference."
        )
    if len(f) == 2 and method == "default":
        correctauto2(
            self,
            f,
            refstrain,
            experiments,
            experimentincludes,
            experimentexcludes,
            conditions,
            conditionincludes,
            conditionexcludes,
            strains,
            strainincludes,
            strainexcludes,
            **options,
        )
    elif len(f) == 1 and method == "default":
        correctauto1(
            self,
            f,
            refstrain,
            experiments,
            experimentincludes,
            experimentexcludes,
            conditions,
            conditionincludes,
            conditionexcludes,
            strains,
            strainincludes,
            strainexcludes,
            **options,
        )
    else:
        print(f"f = {f} must be a list of length 1 or 2.")


def correctauto_checks_l(self, f, method, options, exps, cons, strains):
    """Perform checks on arguents and data before running correctauto."""
    # check for negative fluorescence values and gr is available
    error = False
    for e in exps:
        for c in cons:
            if self.progress["negativevalues"][e]:
                for datatype in f:
                    if (
                        datatype in self.progress["negativevalues"][e]
                        and c in self.progress["negativevalues"][e]
                    ):
                        print(
                            f"{e}: The negative values for {datatype}"
                            f" in {c} will generate NaNs."
                        )
            if "useGPs" in options and options["useGPs"]:
                for s in strains:
                    if f"{s} in {c}" in self.allstrainsconditions[e]:
                        hypers, cvfn = gethypers(self, e, c, s, dtype="gr")
                        if hypers is None or cvfn is None:
                            print(
                                f"You first must run getstats for {s} in {c} "
                                f"for {e}."
                            )
                            error = True
    return error


def correctauto1(
    self,
    f,
    refstrain,
    experiments,
    experimentincludes,
    experimentexcludes,
    conditions,
    conditionincludes,
    conditionexcludes,
    strains,
    strainincludes,
    strainexcludes,
    **kwargs,
):
    """
    Correct autofluorescence for measurements with emissions at one wavelength.

    Corrects for autofluorescence for data with emissions measured at one
    wavelength using the fluorescence of the reference strain
    interpolated to the OD of the tagged strain.

    This method in principle corrects too for the fluorescence of the medium,
    although running correctmedia is still recommended.
    """
    print("Using one fluorescence wavelength.")
    print(f"Correcting autofluorescence using {f[0]}.")
    for e in sunder.getset(
        self,
        experiments,
        experimentincludes,
        experimentexcludes,
        "experiment",
        nonull=True,
    ):
        for c in sunder.getset(
            self,
            conditions,
            conditionincludes,
            conditionexcludes,
            "condition",
            nonull=True,
            nomedia=True,
        ):
            # process reference strain
            if c in self.allconditions[e]:
                refstrfn = processref1(
                    self, f, refstrain, kwargs["figs"], e, c
                )
            else:
                refstrfn = None
            if refstrfn is None:
                continue
            # correct strains
            for s in sunder.getset(
                self,
                strains,
                strainincludes,
                strainexcludes,
                "strain",
                nonull=True,
            ):
                if f"{s} in {c}" in self.allstrainsconditions[e]:
                    t, (od, rawfl) = sunder.extractwells(
                        self.r, self.s, e, c, s, ["OD", f[0]]
                    )
                    # no data
                    if od.size == 0 or rawfl.size == 0:
                        print(f"\n-> No data found for {e}: {s} in {c}.\n")
                        continue
                    # correct autofluorescence for each replicate
                    fl = np.transpose(
                        [
                            rawfl[:, i] - refstrfn(od[:, i])
                            for i in range(od.shape[1])
                        ]
                    )
                    fl[fl < 0] = np.nan
                    if kwargs["useGPs"]:
                        flperod = sample_flperod_with_GPs(
                            self,
                            f[0],
                            t,
                            fl,
                            od,
                            kwargs["flcvfn"],
                            kwargs["bd"],
                            kwargs["nosamples"],
                            e,
                            c,
                            s,
                            kwargs["maxdatapts"],
                            kwargs["figs"],
                        )
                    else:
                        # use only the replicates
                        flperod = np.transpose(
                            [fl[:, i] / od[:, i] for i in range(od.shape[1])]
                        )
                        flperod[flperod < 0] = np.nan
                    # check number of NaN
                    nonans = np.count_nonzero(np.isnan(fl))
                    if np.any(nonans):
                        if nonans == fl.size:
                            print(
                                "\n-> Corrected fluorescence is all NaN "
                                f"for {e}: {s} in {c}.\n"
                            )
                        else:
                            print(
                                f"Warning - {e}: {s} in {c}\n"
                                f"{nonans} corrected data points are"
                                " NaN: the corrected fluorescence"
                                " was negative.",
                            )
                        print("---")
                    # store results
                    autofdict = {
                        "experiment": e,
                        "condition": c,
                        "strain": s,
                        "time": t,
                        f"c{f[0]}": np.nanmean(fl, 1),
                        f"c{f[0]}_err": nanstdzeros2nan(fl, 1),
                        f"c{f[0]}perOD": np.nanmean(flperod, 1),
                        f"c{f[0]}perOD_err": nanstdzeros2nan(flperod, 1),
                    }
                    addtodataframes(self, autofdict)


def processref1(self, f, refstrain, figs, experiment, condition):
    """
    Process reference strain for data with one fluorescence measurement.

    Use lowess to smooth the fluorescence of the reference
    strain as a function of OD.

    Parameters
    ----------
    f: string
        The fluorescence to be corrected. For example, ['mCherry'].
    refstrain: string
        The reference strain. For example, 'WT'.
    figs: boolean
        If True, display fits of the reference strain's fluorescence.
    experiment: string
        The experiment to be corrected.
    condition: string
        The condition to be corrected.

    Returns
    -------
    refstrfn: function
        The reference strain's fluorescence as a function of OD.
    """
    e, c = experiment, condition
    print(f"{e}: Processing reference strain {refstrain} for {f[0]} in {c}.")
    _, (od, fl) = sunder.extractwells(
        self.r, self.s, e, c, refstrain, ["OD", f[0]]
    )
    if od.size == 0 or fl.size == 0:
        raise errors.CorrectAuto(f"{e}: {refstrain} not found in {c}.")
    else:
        odf = od.flatten("F")
        flf = fl.flatten("F")
        if ~np.any(flf[~np.isnan(flf)]):
            return None
        # smooth fluorescence as a function of OD using lowess to minimize
        # refstrain's autofluorescence

        def choosefrac(frac):
            res = lowess(flf, odf, frac=frac)
            refstrfn = interp1d(
                res[:, 0],
                res[:, 1],
                fill_value=(res[0, 1], res[-1, 1]),
                bounds_error=False,
            )
            # max gives smoother fits than mean
            return np.max(np.abs(flf - refstrfn(odf)))

        res = minimize_scalar(choosefrac, bounds=(0.1, 0.99), method="bounded")
        # choose the optimum frac
        frac = res.x if res.success else 0.33
        res = lowess(flf, odf, frac=frac)
        refstrfn = interp1d(
            res[:, 0],
            res[:, 1],
            fill_value=(res[0, 1], res[-1, 1]),
            bounds_error=False,
        )
        if figs:
            # plot fit
            plt.figure()
            plt.plot(odf, flf, ".", alpha=0.5)
            plt.plot(res[:, 0], res[:, 1])
            plt.xlabel("OD")
            plt.ylabel(f[0])
            plt.title(e + ": " + refstrain + " for " + c)
            plt.show(block=False)
        return refstrfn


def correctauto2(
    self,
    f,
    refstrain,
    experiments,
    experimentincludes,
    experimentexcludes,
    conditions,
    conditionincludes,
    conditionexcludes,
    strains,
    strainincludes,
    strainexcludes,
    **kwargs,
):
    """
    Correct autofluorescence for measurements with two emission wavelengths.

    Corrects for autofluorescence using spectral unmixing for data with
    measured emissions at two wavelengths.

    References
    ----------
    CA Lichten, R White, IB Clark, PS Swain (2014). Unmixing of fluorescence
    spectra to resolve quantitative time-series measurements of gene
    expression in plate readers.
    BMC Biotech, 14, 1-11.
    """
    # correct for autofluorescence
    print("Using two fluorescence wavelengths.")
    print(f"Correcting autofluorescence using {f[0]} and {f[1]}.")
    for e in sunder.getset(
        self,
        experiments,
        experimentincludes,
        experimentexcludes,
        "experiment",
        nonull=True,
    ):
        for c in sunder.getset(
            self,
            conditions,
            conditionincludes,
            conditionexcludes,
            labeltype="condition",
            nonull=True,
            nomedia=True,
        ):
            # local r data frame for media corrections
            lrdf = self.r[
                (self.r.experiment == e) & (self.r.condition == c)
            ].copy()
            # correct for background fluorescence using Null strain
            print("Correcting for background fluorescence of media.")
            for fl in f:
                negvalues = find_Null_and_correct(
                    self,
                    lrdf,
                    fl,
                    e,
                    c,
                    kwargs["figs"],
                    kwargs["frac"],
                    kwargs["null_dict"],
                )
                if negvalues:
                    print("Warning: negative values for\n", negvalues)
            # process reference strain
            refqrfn = processref2(
                self, lrdf, f, refstrain, kwargs["figs"], e, c
            )
            if refqrfn is None:
                # too many NaNs in reference strain
                continue
            # process other strains
            for s in sunder.getset(
                self,
                strains,
                strainincludes,
                strainexcludes,
                labeltype="strain",
                nonull=True,
            ):
                if (
                    s != refstrain
                    and f"{s} in {c}" in self.allstrainsconditions[e]
                ):
                    t, (fl_0, fl_1, od) = sunder.extractwells(
                        lrdf, self.s, e, c, s, f.copy() + ["OD"]
                    )
                    if fl_0.size == 0 or fl_1.size == 0:
                        print(f"Warning: No data found for {e}: {s} in {c}.")
                        continue
                    # set negative values to NaNs
                    fl_0[fl_0 < 0] = np.nan
                    fl_1[fl_1 < 0] = np.nan
                    # use mean OD for predicting ra from refstrain
                    odmean = np.nanmean(od, axis=1)
                    # correct autofluorescence
                    ra = refqrfn(odmean)
                    fl = applyautoflcorrection(self, ra, fl_0, fl_1)
                    fl[fl < 0] = np.nan
                    if kwargs["useGPs"]:
                        # sample to estimate errors
                        flperod = sample_flperod_with_GPs(
                            self,
                            f[0],
                            t,
                            fl,
                            od,
                            kwargs["flcvfn"],
                            kwargs["bd"],
                            kwargs["nosamples"],
                            e,
                            c,
                            s,
                            kwargs["maxdatapts"],
                            kwargs["figs"],
                        )
                    else:
                        # use the replicates
                        flperod = fl / od
                        flperod[flperod < 0] = np.nan
                    # store results
                    autofdict = {
                        "experiment": e,
                        "condition": c,
                        "strain": s,
                        "time": t,
                        f"c{f[0]}": np.nanmean(fl, 1),
                        f"c{f[0]}_err": naniqrzeros2nan(fl, 1),
                        f"c{f[0]}perOD_err": naniqrzeros2nan(flperod, 1),
                        f"c{f[0]}perOD": np.nanmean(flperod, 1),
                    }
                    # add to data frames
                    addtodataframes(self, autofdict)
                    print("---")


def processref2(self, lrdf, f, refstrain, figs, experiment, condition):
    """
    Process reference strain for spectral unmixing.

    Requires data with two fluorescence measurements.

    Use lowess to smooth the ratio of emitted fluorescence measurements
    so that the reference strain's data is corrected to zero as best
    as possible.

    Parameters
    ----------
    lrdf: pd.DataFrame
        A copy of the r data frame with fluorescence corrected for media.
    f: list of strings
        The fluorescence measurements. For example, ['GFP', 'AutoFL'].
    refstrain: string
        The reference strain. For example, 'WT'.
    figs: boolean
        If True, display fits of the fluorescence ratios.
    experiment: string
        The experiment to be corrected.
    condition: string
        The condition to be corrected.

    Returns
    -------
    qrfn: function
        The ratio of the two fluorescence values for the reference strain
        as a function of OD.
    """
    e, c = experiment, condition
    print(f"{e}: Processing reference strain {refstrain} for {f[0]} in {c}.")
    # refstrain data
    t, (f0, f1, od) = sunder.extractwells(
        lrdf, self.s, e, c, refstrain, f.copy() + ["OD"]
    )
    if f0.size == 0 or f1.size == 0 or od.size == 0:
        raise errors.CorrectAuto(f"{e}: {refstrain} not found in {c}.")
    else:
        f0[f0 < 0] = np.nan
        f1[f1 < 0] = np.nan
        odf = od.flatten("F")
        odrefmean = np.mean(od, 1)
        qrf = (f1 / f0).flatten("F")
        if np.all(np.isnan(qrf)):
            print(f"{e}: {refstrain} in {c} has too many NaNs.")
            return
        # smooth to minimise autofluorescence in refstrain

        def choosefrac(frac):
            qrfn, _ = find_qrfn(qrf, odf, frac)
            flref = applyautoflcorrection(self, qrfn(odrefmean), f0, f1)
            return np.max(np.abs(flref))

        res = minimize_scalar(choosefrac, bounds=(0.1, 0.99), method="bounded")
        # calculate the relationship between qr and OD
        frac = res.x if res.success else 0.95
        # apply lowess and find qrfn
        qrfn, res = find_qrfn(qrf, odf, frac)
        if figs:
            plt.figure()
            plt.plot(odf, qrf, ".", alpha=0.5)
            plt.plot(res[:, 0], res[:, 1])
            plt.xlabel("OD")
            plt.ylabel(f[1] + "/" + f[0])
            plt.title(e + ": " + refstrain + " in " + c)
            plt.show(block=False)
        # check autofluorescence correction for reference strain
        flref = applyautoflcorrection(self, qrfn(odrefmean), f0, f1)
        flrefperod = flref / od
        # set negative values to NaNs
        flref[flref < 0] = np.nan
        flrefperod[flrefperod < 0] = np.nan
        # store results
        autofdict = {
            "experiment": e,
            "condition": c,
            "strain": refstrain,
            "time": t,
            f"c{f[0]}": np.nanmean(flref, 1),
            f"c{f[0]}perOD": np.nanmean(flrefperod, 1),
            f"c{f[0]}_err": nanstdzeros2nan(flref, 1),
            f"c{f[0]}perOD_err": nanstdzeros2nan(flrefperod, 1),
        }
        addtodataframes(self, autofdict)
        return qrfn


def find_qrfn(qrf, odf, frac):
    """Use lowess and then interpolation to find qrfn."""
    res = lowess(qrf, odf, frac)
    qrfn = interp1d(
        res[:, 0],
        res[:, 1],
        fill_value=(res[0, 1], res[-1, 1]),
        bounds_error=False,
    )
    return qrfn, res


def applyautoflcorrection(self, ra, f0data, f1data):
    """Correct for autofluorescence returning an array of replicates."""
    nr = f0data.shape[1]
    raa = np.reshape(np.tile(ra, nr), (np.size(ra), nr), order="F")
    return (raa * f0data - f1data) / (
        raa - self._gamma * np.ones(np.shape(raa))
    )


def addtodataframes(self, datadict):
    """Added dict of data to s data frame."""
    newdf = pd.DataFrame(datadict)
    key_cols = ["experiment", "condition", "strain", "time"]
    self.s = gu.merge_df_into(self.s, newdf, key_cols)


def nanstdzeros2nan(a, axis=None):
    """Like nanstd but setting zeros to nan."""
    err = np.nanstd(a, axis)
    err[err == 0] = np.nan
    return err


def naniqrzeros2nan(a, axis=None):
    """Interquartile range but setting zeros to nan."""
    iqr = np.nanquantile(a, 0.75, axis) - np.nanquantile(a, 0.25, axis)
    iqr[iqr == 0] = np.nan
    return iqr


def gethypers(self, exp, con, s, dtype="gr"):
    """Find parameters for GP from sc data frame."""
    sdf = self.sc[
        (self.sc.experiment == exp)
        & (self.sc.condition == con)
        & (self.sc.strain == s)
    ]
    if sdf.empty:
        return None, None
    else:
        try:
            cvfn = sdf[f"gp_for_{dtype}"].values[0]
            hypers = [
                sdf[col].values[0]
                for col in sorted(sdf.columns)
                if ("hyper" in col and dtype in col)
            ]
            if np.any(np.isnan(hypers)):
                return None, None
            else:
                return hypers, cvfn
        except KeyError:
            return None, None


def instantiateGP(hypers, cvfn, x, y, maxdatapts):
    """Instantiate a Gaussian process."""
    xa, ya, _, _ = preprocess_data(x, y, merrors=[], maxdatapts=maxdatapts)
    # bounds are irrelevant because parameters are optimal
    go = getattr(gp, cvfn + "GP")(
        {0: (-5, 5), 1: (-4, 4), 2: (-5, 2)}, xa, ya, warnings=False
    )
    go.lth_opt = hypers
    go.success = True
    # make predictions so that samples can be generated
    go.predict(x, derivs=2)
    return go


def sample_ODs_with_GP(
    self, experiment, condition, strain, t, od, nosamples, maxdatapts
):
    """Instantiate Gaussian process for log OD and sample."""
    hypers, cvfn = gethypers(self, experiment, condition, strain, dtype="gr")
    if hypers is None or cvfn is None:
        raise SystemExit(
            f"You first must run getstats for {strain} in {condition} "
            f"for {experiment} unless useGPs=False."
        )
    od[od < 0] = np.nan
    # initialise GP for log ODs
    go = instantiateGP(hypers, cvfn, t, np.log(od), maxdatapts)
    # sample
    lod_samples = go.sample(nosamples)
    return lod_samples


def sample_flperod_with_GPs(
    self,
    flname,
    t,
    fl,
    od,
    flcvfn,
    bd,
    nosamples,
    e,
    c,
    s,
    maxdatapts=None,
    figs=True,
    logs=True,
    negs2nan=True,
):
    """
    Generate samples of fluorescence per OD.

    Smooth and sample fluorescence using a Gaussian process.
    Sample ODs using the Gaussian process generated by getstats.
    """
    if np.any(fl[~np.isnan(fl)]):
        # run GP for fluorescence or log fluorescence
        # omfitderiv deals with NaNs
        if logs:
            fitvar = f"log_{flname}"
        else:
            fitvar = flname
        f_fitderiv = runfitderiv(
            t,
            fl,
            fitvar,
            f"d/dt_{fitvar}",
            experiment=e,
            condition=c,
            strain=s,
            bd=bd,
            cvfn=flcvfn,
            logs=logs,
            negs2nan=negs2nan,
            maxdatapts=maxdatapts,
            plotlocalmax=False,
        )
        if not f_fitderiv.success:
            print(f"-> Fitting fluorescence failed for {e}: {s} in {c}.")
            return np.nan * np.ones((t.size, nosamples))
        # samples
        lod_samples = sample_ODs_with_GP(
            self, e, c, s, t, od, nosamples, maxdatapts
        )
        f_samples = f_fitderiv.fitderivsample(nosamples)[0]
        if logs:
            flperod_samples = np.exp(f_samples - lod_samples)
        else:
            flperod_samples = f_samples * np.exp(-lod_samples)
    else:
        print("No positive data.")
        # all NaN
        flperod_samples = np.nan * np.ones((t.size, nosamples))
    return flperod_samples
